// VinciaShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Antenna class 
  
#ifndef VINCIA_VinciaAntenna_H
#define VINCIA_VinciaAntenna_H

#include "Basics.h"
#include "Event.h"
#include "PythiaStdlib.h"
#include "VinciaCommon.h"

namespace Vincia {
  
  // The Antenna class: contains a generic antenna function

  class Antenna {
    
  public: 
    //Bare constructor: name only
  Antenna(string Name="antenna") : name(Name),
      humanName(Name) {useFiniteTerms = true; isInitPtr = false;};
    
    //Constructor with PDG codes: 
    //  Antenna(string Name,vector<int> IdOld, vector<int> IdNew)
    //  : name(Name), idOld(IdOld), idNew(IdNew), 
    //	isInit(false), {useFiniteTerms=true;};
    
    // Destructor.
    virtual ~Antenna() {};  
    
    // Initialize
    bool init();
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
    bool setC(int i=0,int j=0,double c=0.0); // Set C[i][j]
    bool setD(int j=0,double d=0.0);         // Set D[j]
    bool setE(int j=0,double e=0.0);         // Set E[j]
    bool setK(double k=0.0);                 // Set K
    bool setParm(bool&, string);
    bool setParm(int&, string);
    bool setParm(double&, string);
    bool setParm(string&, string);
    void setHumanName(string humanNameIn) {humanName = humanNameIn;}
    
    // Functions to "get"
    string getHumanName() { return humanName; }
    string getName() { return name; }
    int getInputMode() {return inputMode;}
    int getEvolutionType() {return evolutionType;};     
    int getKineMapType() {return kineMapType;}  
    double getChargeFactor() {return chargeFactor;}  
    int getCutoffType() { return cutoffType; }                 
    double getCutoffScale() { return cutoffScale; }                 
    
    // The antenna function for this Sudakov
    double a2(double sa1,double s1b,double s) {
      return (s != 0) ? a2(sa1/s,s1b/s) : 0.0;};
    double a2(double,double); // ya1,y1b
    
    // The finite part polynomial coefficients 
    // (cannot be made static)
    void readC();                      // Read C,D,E,K from Settings
    double getC(int i=0, int j=0);     // Return C[i][j]
    double getD(int j=0);              // Return D[j]
    double getE(int j=0);              // Return E[j]
    double getK();                     // Return K
    void clear();                      // Clear all C,D,E,K
    
    // Print a label for this antenna
    void label(ostream& os=cout);
    
  private:
    
    // Antenna Parameters 
    int evolutionType;
    int kineMapType;
    double chargeFactor;
    
    // Cutoff
    int cutoffType;
    double cutoffScale;

    // Antenna function name 
    string name;       // in xml/cmnd files (e.g., Vincia:qq:Emit)
    string humanName;  // human readable format, derived.
    
    // In- and Out- particles in PDG code format
    vector<int> idOld;
    vector<int> idNew;
    vector<string> nameOld;
    vector<string> nameNew;
    
    // Coefficients of the antenna function 
    // (There is only one possible input mode for now: Laurent)
    int inputMode;  
    // Coefficients of Laurent series
    map<int, map<int, double> > LaurentC;
    // Additional coefficients for sector antennae
    map<int,double> LaurentD,LaurentE;
    double LaurentK;
    // Flag whether to use finite terms or not
    bool useFiniteTerms;
    // Dead Zone Avoidance Factor
    double deadZoneAvoidanceFactor;
    
    // Initialized flag
    bool isInit;
    
    // Pointers to Pythia-8 classes
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
    // Verbose setting
    int verbose;

  };

  // AntennaSet = List of Antennae 
  
  class AntennaSet {

  public:
    
    void setAnt(int iAntIn, Antenna antIn) {
      if (iAntIn >= int(antSav.size())) antSav.resize(iAntIn+1);
      antSav[iAntIn]=antIn;
      antIndxSav[antIn.getName()]=iAntIn;
    }
    bool exists(int iAntIn) {
      return (iAntIn<0 || iAntIn >= int(antSav.size())) ? false : true;
    }
    Antenna* getAnt(int iAntIn) {
      return (exists(iAntIn) ? &antSav[iAntIn] : 0);
    }
    string getName(int iAntIn) {
      return (exists(iAntIn) ? antSav[iAntIn].getName() : "none");
    }
    string getHumanName(int iAntIn) {
      return (exists(iAntIn) ? antSav[iAntIn].getHumanName() : "none");
    }
    int getIndx(string antNameIn) {return antIndxSav[antNameIn];}
    
  private:
    
    vector<Antenna> antSav;
    map<string,int> antIndxSav;

  }; // End class AntennaSet
  
  
} // End Vincia namespace

#endif
