// VinciaAntenna.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the VinciaAntenna header file
#include "VinciaAntenna.h"

using namespace Pythia8;

namespace Vincia {
  
  //**********************************************************************
  
  // The Antenna class and its members
  
  //*********
  
  // Initialize Antenna

  bool Antenna::init() {

    // Check that pointers initialized
    if (! isInitPtr) {
      printOut("Antenna::init","Cannot initialize, pointers not set.");
      return false;
    }

    // Innocent until proven guilty
    isInit=true;

    // Force inputMode = 1 (double Laurent rep) for the time being
    inputMode = 1;

    // Set verbose level
    setParm(verbose,"verbose");

    // Read generic in- and out- states
    string o1,o2,n1,n2,n3,n4;
    if (name.find("Vincia:qQ:emit") != string::npos ) {
      o1="q";o2="qbar";n1="q";n2="g";n3="qbar";
    }
    else if (name.find("Vincia:qg:emit") != string::npos ) {
      o1="q";o2="g";n1="q";n2="g";n3="g";
    }
    else if (name.find("Vincia:gg:emit") != string::npos ) {
      o1="g";o2="g";n1="g";n2="g";n3="g";}      
    else if (name.find("Vincia:qg:split") != string::npos ) {
      o1="q";o2="g";n1="q";n2="qbar'";n3="q'";
    }
    else if (name.find("Vincia:gg:split") != string::npos ) {
      o1="g";o2="g";n1="g";n2="qbar";n3="q";
    }
    else {
      if (verbose >= 1)
	printErr(name+":init","unknown generic antenna");
      return false;
    }
    nameOld.push_back(o1);
    nameOld.push_back(o2);
    nameNew.push_back(n1);
    nameNew.push_back(n2);
    nameNew.push_back(n3);
    if (n4 != "") nameNew.push_back(n4);
    for (int i=0;i<int(nameOld.size());i++) {
      if (nameOld[i]=="q") idOld.push_back(1);
      if (nameOld[i]=="qbar") idOld.push_back(-1);
      if (nameOld[i]=="g") idOld.push_back(21);
      if (nameOld[i]=="gamma") idOld.push_back(22);
    }
    for (int i=0;i<int(nameNew.size());i++) {
      if (nameNew[i]=="q") idNew.push_back(1);
      if (nameNew[i]=="qbar") idNew.push_back(-1);
      if (nameNew[i]=="g") idNew.push_back(21);
      if (nameNew[i]=="gamma") idNew.push_back(22);
      if (nameNew[i]=="q'") idNew.push_back(2);
      if (nameNew[i]=="qbar'") idNew.push_back(-2);
    }
    if (humanName == name) humanName=name.substr(7,name.length());

    // First check whether antenna is 2->3
    if (nameOld.size() != 2 || nameNew.size() != 3) {
      if (verbose >= 1)
	printErr(name+":init","unknown antenna type: "+
		 num2str(int(idOld.size()))+"->"+num2str(int(idNew.size())));
      return false;
    }

    // Check if switched off
    if ( ! settingsPtr->flag(name)) {
        clear();
	return true;
    }

    // Set whether finite terms to be used in evolution
    isInit=isInit&&setParm(useFiniteTerms,"useFiniteTerms");
    // Initialize parameter to use when checking for dead zones
    isInit=isInit&&setParm(deadZoneAvoidanceFactor,"deadZoneAvoidanceFactor");
    
    // Check charge conservation
    double sumQ=0.0;
    for (unsigned int i=0;i<idOld.size();i++) 
      sumQ+=particleDataPtr->charge(idOld[i]);
    for (unsigned int i=0;i<idNew.size();i++) 
      sumQ-=particleDataPtr->charge(idNew[i]);
    if (abs(sumQ) > TINY) {
      if (verbose >= 1)
	printErr(name+":init","antenna does not conserve charge");
      cout<<"   sumQ = "<<sumQ<<" Antenna ID :"<<idOld[1]<<" "
	  <<idOld[2]<<" -> "
	  <<idNew[1]<<" "<<idNew[2]<<" "<<idNew[3]<<endl;
      return false; 
    }

    // Read in antenna function terms (zero if switched off)    
    readC();  

    // Check that antenna function is positive definite 
    // (if not, try to repair by increasing constant term)
    // Also check that finite terms do not introduce dead zones
    bool isPos(true);
    bool noDead(true);
    //    int nPoints = (verbose > 1) ? min(1e4,1.0/sqrt(TINY)) : 100;
    int nPointsCheck = 100;

    for (int i=1;i<nPointsCheck;i++) {
      
      // Generate a random point in PS triangle
      double y12=rndmPtr->flat()*1.0;
      double y23=rndmPtr->flat()*1.0;
      if (y12+y23>1.0) { y12=1.0-y12; y23=1.0-y23; }

      // Evaluate a2 with and without finite terms
      double aFull   = a2(y12,y23);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = a2(y12,y23);
      useFiniteTerms = useFinSav;

      // Check if overall function positive definite
      double c00(getC(0,0));
      if (c00 > 50) {
	if (verbose >= 0) 
	  printErr(name+":init",
		   "error: no positivity even with LaurentC(0,0) = 50.0");
	return false;
      }
      if (aFull < -1.0*TINY) {
	if (verbose >= 8)
	  printOut(name+":init","a = "+num2str(aFull)
		   +" < 0 at y12,y23 = "+num2str(y12)+","+num2str(y23));
	c00=max(getC(0,0)+0.5,-10.0);
	setC(0,0,c00);
	isPos=false;
	i=1;
      }

      // Check if finite terms introduce dead zones
      if (aFull < deadZoneAvoidanceFactor * aSing) {
	c00=max(getC(0,0)+0.1,-10.0);
	setC(0,0,c00);
	noDead=false;
	i=1;
      }

    }
    if (not isPos) {
      if (verbose >= 1)
	printErr(humanName+":init","warning: LaurentC(0,0) changed to "+
		 num2str(getC(0,0))+" to ensure positivity");
    } else if (not noDead) {
      if (verbose >= 1)
	printErr(humanName+":init","warning: LaurentC(0,0) changed to "+
		 num2str(getC(0,0))+" to ensure no dead zones");
    } else {
      if (verbose >= 4)
	printOut(humanName+":init","ok: positive definite, no dead zones");
    }      

    // Initialize evolution settings
    isInit=isInit&&setParm(evolutionType,"evolutionType"); 
    isInit=isInit&&setParm(kineMapType,"kineMapType");

    // Set chargeFactor (i.e., CA, CF, TR, etc.)
    isInit=isInit&&setParm(chargeFactor,"chargeFactor");

    // IR hadronization cutoff type and scale
    isInit=isInit&&setParm(cutoffType,"cutoffType");
    if (cutoffType == 0) cutoffType=evolutionType;
    isInit=isInit&&setParm(cutoffScale,"cutoffScale");
    
    // Return true or false, depending on how checks went.
    return isInit;
  }
  
//*********

// Set antenna parameters from Settings database

bool Antenna::setParm(bool& par, string parName) {
  bool found=true;
  if (settingsPtr->isFlag(name+":"+parName)) 
    par = settingsPtr->flag(name+":"+parName);
  else if (settingsPtr->isFlag("Vincia:"+parName)) 
    par = settingsPtr->flag("Vincia:"+parName);
  else if (settingsPtr->isFlag("Vincia:"+parName)) 
    par = settingsPtr->flag("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName); }
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found"); }
  return found;
}

bool Antenna::setParm(int& par, string parName) {
  bool found=true;
  if (settingsPtr->isMode(name+":"+parName)) 
    par = settingsPtr->mode(name+":"+parName);
  else if (settingsPtr->isMode("Vincia:Shower:"+parName)) 
    par = settingsPtr->mode("Vincia:Shower:"+parName);
  else if (settingsPtr->isMode("Vincia:"+parName)) 
    par = settingsPtr->mode("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose>= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(double& par, string parName) {
  bool found=true;
  if (settingsPtr->isParm(name+":"+parName)) 
    par = settingsPtr->parm(name+":"+parName);
  else if (settingsPtr->isParm("Vincia:Shower:"+parName)) 
    par = settingsPtr->parm("Vincia:Shower:"+parName);
  else if (settingsPtr->isParm("Vincia:"+parName)) 
    par = settingsPtr->parm("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(string& par, string parName) {
  bool found=true;
  if (settingsPtr->isWord(name+":"+parName)) 
    par = settingsPtr->word(name+":"+parName);
  else if (settingsPtr->isWord("Vincia:Shower:"+parName)) 
    par = settingsPtr->word("Vincia:Shower:"+parName);
  else if (settingsPtr->isWord("Vincia:"+parName)) 
    par = settingsPtr->word("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

//*********

// Read in Laurent coefficients from Settings database

void Antenna::readC() {
  if (verbose >= 5)
    printOut(humanName+":readC","reading antenna coefficients ");
  // Antenna function
  clear();
  if (inputMode == 1) {
    // Input as Laurent Series
    for (int i=-1;i<=LaurentIndexMax;i++) {
      string baseName("LaurentC("+num2str(i,0)+",");
      for (int j=-1;j<=LaurentIndexMax;j++) {
	string CName(baseName+num2str(j,0)+")"); 
	if (settingsPtr->isParm(name+":"+CName)) {	
	  if (setC(i,j,settingsPtr->parm(name+":"+CName))) {
	    int verb = (getC(i,j) == 0.0) ? 8 : 6; 
	    if (verbose >= verb)
	      printOut(humanName+":readC",
		     CName+" = "+num2str(getC(i,j),6));
	  }
	} 
      } 
      string DName("LaurentD("+num2str(i,0)+")");
      if (settingsPtr->isParm(name+":"+DName)) {	
	if (setD(i,settingsPtr->parm(name+":"+DName))) {
	  int verb = (getD(i) == 0.0) ? 8 : 6; 
	  if (verbose >= verb)
	    printOut(humanName+":readC",
		   DName+" = "+num2str(getD(i),6));
	}
      } 
      string EName("LaurentE("+num2str(i,0)+")");
      if (settingsPtr->isParm(name+":"+EName)) {	
	if (setE(i,settingsPtr->parm(name+":"+EName))) {
	  int verb = (getE(i) == 0.0) ? 8 : 6; 
	  if (verbose >= verb)
	    printOut(humanName+":readC",
		     EName+" = "+num2str(getE(i),6));
	}
      } 
    } 
    string KName("LaurentK");
    if (settingsPtr->isParm(name+":"+KName)) {	
      if (setK(settingsPtr->parm(name+":"+KName))) {
	int verb = (getK() == 0.0) ? 8 : 6; 
	if (verbose >= verb)
	  printOut(humanName+":readC",
		   KName+" = "+num2str(getK(),6));
      }
    } 
  } else {
    // Enter alternative input forms here
  }
}
  
//*********

// Set a Laurent coefficient directly

bool Antenna::setC(int i,int j, double cIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (i < -1 or j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setC",
		 "leading term is (-1,-1) : ignoring entry ("
		 +num2str(i)+","+num2str(j)+")");
      return false;
    }
    else if (cIn == 0) {
      if (LaurentC.find(i) != LaurentC.end()) {
	if (LaurentC[i].find(j) != LaurentC[i].end()) LaurentC[i].erase(j);
      } 
    }
    else if (i == -1 && j == -1 && cIn>0) LaurentC[i][j]=cIn;
    else if (j == -1 || i == -1) LaurentC[i][j]=cIn;
    else LaurentC[i][j]=cIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1)
      printErr(humanName+":setC",
	     "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setD(int j, double dIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setD",
	       "leading term is (-1) : ignoring entry ("
	       +num2str(j)+")");
      return false;
    }
    else if (dIn == 0) {
      if (LaurentD.find(j) != LaurentD.end()) LaurentD.erase(j);
    }
    else LaurentD[j]=dIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setD",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setE(int j, double eIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setE",
		 "leading term is (-1) : ignoring entry ("
		 +num2str(j)+")");
      return false;
    }
    else if (eIn == 0) {
      if (LaurentE.find(j) != LaurentE.end()) LaurentE.erase(j);
    }
    else LaurentE[j]=eIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setE",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setK(double kIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    LaurentK=kIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setK",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}
  
//*********

// Function to "get" value of a Laurent coefficient

double Antenna::getC(int i,int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    it=LaurentC.find(i);
    if (it==LaurentC.end()) return 0.0;
    jt=LaurentC[i].find(j);
    if (jt==LaurentC[i].end()) return 0.0;
    return LaurentC[i][j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getD(int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,double>::iterator jt;
    jt=LaurentD.find(j);
    if (jt==LaurentD.end()) return 0.0;
    return LaurentD[j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getE(int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,double>::iterator jt;
    jt=LaurentE.find(j);
    if (jt==LaurentE.end()) return 0.0;
    return LaurentE[j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getK() {       
  if (inputMode == 1) {
    return LaurentK;
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
  
//*********

// Zero all Laurent coefficients

void Antenna::clear() { 
  if (inputMode == 1) {
    // Clear Laurent series coeffiecents
    LaurentC.clear(); 
    LaurentD.clear(); 
    LaurentE.clear(); 
    LaurentK=0.0;
  } else {
    // Enter alternative input modes here
  }
}

//*********

// The antenna function 

double Antenna::a2(double y12, double y23) {

  double val=0.0;

  // Input as Laurent series
  if (inputMode == 1) {
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    // Add C(i,j) terms
    for (it=LaurentC.begin();it!=LaurentC.end();++it) {
      int i=it->first;
      for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
	int j=jt->first;  
	// Only add finite terms if they are "on"
	if (i < 0 || j < 0 || useFiniteTerms) {
	  double c=jt->second;      
	  val += c * pow(y12,i) * pow(y23,j);
	}
      }      
    }
    // Add D(j) terms
    for (jt=LaurentD.begin();jt!=LaurentD.end();++jt) {
      int j=jt->first;
      double d=jt->second;
      val += d * pow(y23,j) / (1-y12) ;
    }
    // Add E(j) terms
    for (jt=LaurentE.begin();jt!=LaurentE.end();++jt) {
      int j=jt->first;
      double e=jt->second;
      val += e * pow(y12,j) / (1-y23) ;
    }
    // Add K term
    if (LaurentK != 0) {
      val += LaurentK / (1-y12) / (1-y23) ;
    }
  } else {
    // Enter alternative input forms here
  }

  // Return val
  return val;
}   

} // End Vincia Namespace
  
