// VinciaTools.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for VINCIA's Toolbox.

#include "VinciaTools.h"

// Declare the F77 3->2 clustering map for use below
extern"C" {
  void vc3to2_(int *MMAP, double PA[4], double PR[4], double PB[4],
	       double PAHAT[4], double PBHAT[4], int *IFAIL);
}

namespace Vincia {

//***********************************************************************

// VinClu: VINCIA Clustering 
// A set of 3->2 dipole-antenna mappings

// ******

// vc3to2: main clustering map
// Input:
//   iMap    = map number ( 1 : ARIADNE, 2 : PS, 3 : Kosower)
//   pIn     = Vec4 list (arbitrary length, but greater than 3)
//   a,r,b   = indices of 3 particles in pIn to be clustered
// Output:  
//   pClu    = pIn but with the a and b momenta replaced by the clustered
//             aHat and bHat and with r erased.
// I.e.,
//   pIn(...,a,...,r-1,r,r+1,...,b,...) 
//     -> 
//   pOut(...,aHat,...,r-1,r+1,...,bHat,...)
//   with {a,r,b} -> {aHat,bHat} using kinematics map iMap

  void VinClu::vc3to2(int iMap,vector<Vec4> pIn,int a, int r, int b, 
		      vector<Vec4>& pClu) {

    // Intialize and sanity check
    pClu=pIn;
    if (max(max(a,r),b) > int(pIn.size()) || min(min(a,r),b) < 0) return;
    
    //cout<<" vinclu: mar = "<<m(pIn[a],pIn[r])<<" mrb = "<<m(pIn[r],pIn[b])<<endl;

    // Use fortran map for ARIADNE and PS maps
    if (iMap == 1 || iMap == 2) {
      // Initialize variables for use in re-clusterings
      double pa[4],pr[4],pb[4],pahat[4],pbhat[4];
      int ifail=0;      

      // Do reclustering
      Vec4toD4(pIn[a],pa);
      Vec4toD4(pIn[r],pr);
      Vec4toD4(pIn[b],pb);
      /*
      cout<<" pa = "<<pa[0]<<" "<<pa[1]<<" "<<pa[2]<<" "<<pa[3]<<endl;
      cout<<" pr = "<<pr[0]<<" "<<pr[1]<<" "<<pr[2]<<" "<<pr[3]<<endl;
      cout<<" pb = "<<pb[0]<<" "<<pb[1]<<" "<<pb[2]<<" "<<pb[3]<<endl;
      */
      vc3to2_(&iMap,pa,pr,pb,pahat,pbhat,&ifail);
      
      // Check return code
      if (ifail != 0) return;
      
      // Store reclustered momenta
      pClu[a]=(Vec4(pahat[0],pahat[1],pahat[2],pahat[3]));
      pClu[b]=(Vec4(pbhat[0],pbhat[1],pbhat[2],pbhat[3]));
      pClu.erase(pClu.begin()+r);
      
    } 
    
    // Use C++ map for David's map 
    else if (iMap == 3) {
      ks3to2(pIn[a],pIn[r],pIn[b],pClu[a],pClu[b]);
      pClu.erase(pClu.begin()+r);
    }
    
  // Probabilistic PS map
  // The current setting does improve slightly on the type 2 map,
  // but not enough to warrant being included as an available 
  // possibility in the VINCIA shower.
  // A or B as emitter selected by collinear splitting function
  else {

    // Compute invariants
    double sij  = m2(pIn[a],pIn[r]);
    double sjk  = m2(pIn[r],pIn[b]);
    // double sik  = m2(pIn[a],pIn[b]);
    // double sijk = m2(pIn[a],pIn[r],pIn[b]);
    
    // Compute probability for A to be the emitter
    // (here based on the collinear terms only. Dividing the soft term
    //  equally on the two possibilities gave a worse performance in the 
    //  matrix-element comparisons, indistinguishable from that of the 
    //  type 2 map)
    // double soft = 2.*sik*sijk/sij/sjk;
    double colA = sjk/sij;
    double colB = sij/sjk;
    double PA   = colA / (colA + colB);
    
    // Special maps: force A or B to be emitter
    if (iMap == -1) PA = 1.0;
    else if (iMap == -2) PA = 0.0;

    // Initialize variables for use in re-clusterings
    double pa[4],pr[4],pb[4],pahat[4],pbhat[4];
    int ifail=0;
  
    // Store momenta for reclustering
    Vec4toD4(pIn[a],pa);
    Vec4toD4(pIn[r],pr);
    Vec4toD4(pIn[b],pb);

    // If B is emitter, A recoils longitudinally
    if (rndmPtr->flat() > PA) {
      
      // Do reclustering, force B as emitter
      int thisMap = -2;
      vc3to2_(&thisMap,pa,pr,pb,pahat,pbhat,&ifail);
    
    }

    // If A is emitter, B recoils longitudinally
    else {

      // Do reclustering, force A as emitter
      int thisMap = -1;
      vc3to2_(&thisMap,pa,pr,pb,pahat,pbhat,&ifail);
    
    }
    
    // Check return code
    if (ifail != 0) return;
  
    // Store reclustered momenta
    pClu[a]=(Vec4(pahat[0],pahat[1],pahat[2],pahat[3]));
    pClu[b]=(Vec4(pbhat[0],pbhat[1],pbhat[2],pbhat[3]));
    pClu.erase(pClu.begin()+r);

  } 

}

// ******

// Kosower's 3to2 clustering map (used by GGG in 4-parton functions)

  void VinClu::ks3to2(Vec4 pi, Vec4 pj, Vec4 pk, Vec4& pOutI, Vec4& pOutK) {

    // Compute invariants
    double sij  = m2(pi,pj);
    double sjk  = m2(pj,pk);
    double sik  = m2(pi,pk);
    double sijk = m2(pi,pj,pk);
    
    // Compute coefficients (choice of r fixed to the one used by GGG)
    double r    = sjk / (sij + sjk); 
    double rho  = sqrt(1.0+(4*r*(1.0-r)*sij*sjk)/sijk/sik);
    double x    = 0.5/(sij+sik)*((1+rho)*sijk-2*r*sjk);
    double z    = 0.5/(sjk+sik)*((1-rho)*sijk-2*r*sij);
    
    // Compute reclustered momenta
    pOutI       =     x*pi +     r*pj +     z*pk;
    pOutK       = (1-x)*pi + (1-r)*pj + (1-z)*pk;
    
  }
  
  //**********************************************************************
  
  // Rambo phase space generator
  // p[0][i] is the energy of the i'th parton (out of N, with ECM^2 = S)
  
  double Rambo::genPoint(double S,int N,double p[4][8])
  {
    double* R=new double[4];for (int j=0;j<4;j++) R[j]=0.0;
    for (int i=0;i<N;i++) {
      double c=-1.0+2.0*rndmPtr->flat();
      double s=sqrt(1.0-c*c);
      double f=2.0*M_PI*rndmPtr->flat();
      double r12=0.0;
      while (r12==0.0) {
	double r1=rndmPtr->flat();
	double r2=rndmPtr->flat();
	r12=r1*r2; }
      double En=-log(r12);
      p[0][i]=En;
      p[3][i]=En*c;
      p[2][i]=En*s*cos(f);
      p[1][i]=En*s*sin(f);
      for (int j=0;j<4;j++) R[j]+=p[j][i];
    }
    double Rmass=sqrt(R[0]*R[0]-R[1]*R[1]-R[2]*R[2]-R[3]*R[3]);
    for (int j=0;j<4;j++) R[j]/=-Rmass;
    double a=1.0/(1.0-R[0]);
    double x=sqrt(S)/Rmass;
    for (int i=0;i<N;i++) {
      double bq=R[1]*p[1][i]+R[2]*p[2][i]+R[3]*p[3][i];
      for (int j=1;j<4;j++)
	p[j][i]=x*(p[j][i]+R[j]*(p[0][i]+a*bq));
      p[0][i]=x*(-R[0]*p[0][i]+bq);
    }
    delete R;
    return 1.0;
  }

}
